<?php
/**
 * @file transport.remote_entity.file.inc
 * Contains the remote entity class, interfaces, and hooks for file.
 *
 * Files are a pseudoentity -- that is, Drupal 6 core does not treat them
 * as such and there is no file module.
 * Where files need to interact with other entities, use system module as
 * the hook provider.
 */

/**
 * Entity class for files.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 *
 * WARNING: some file API functions expect files to be arrays of data, but 
 * within the Remote entity handler we're storing an object in $handler->entity,
 * hence this will require casting to array in some cases.
 *
 * Files are only retrieved once; if a file exists locally then it is skipped.
 * (This is based on the remote fid, rather than the filename which could
 * merely be a coincidence.)
 * This is because Drupal core and filefield never change
 * a file or its {files} table record once it is saved, AFAIK.
 * (Please file an issue if there are special cases!)
 */
class TransportRemoteEntity_file extends TransportRemoteEntity {
  /**
   * The name of the property that holds the entity's id.
   */
  public $primary_key = 'fid';

  /**
   * Phase handler: file - init.
   */
  function run_phase_init() {
    // It is extremely important to call the parent method here.
    parent::run_phase_init();
    
    // Skip if this file has already been retrieved.
    if ($this->local_exists) {
      $this->advance_phase = FALSE;
      $this->status = 'skip';
      //dsm($this, 'skip file');
    }
  }
  
  /**
   * Phase handler: file - local_load.
   *
   * We technically never do anything here anyway because existing files are
   * skipped, but for completion we're overriding this so local files are not
   * loaded.
   */
  function run_phase_local_load() {
  }

  /**
   * Phase handler: file - dependencies.
   *
   * Files don't have dependencies, at least not on D5.
   * The uid in theory should already have been brought over -- we hope!
   * Arg that means we may need to version this too????
   */  
  function run_phase_dependencies() {
  }
  
  /**
   * Phase handler: file - translate.
   */  
  function run_phase_translate() {
    // TODO: translate file uid property.
  }
  
  /**
   * Phase handler: file - save.
   *
   * Oh argh. This may have to be versioned for files :(
   * Which means phases are versioned inconsistently per type... fun!
   */  
  function run_phase_save() {
    //dsm($this, 'phase save file');
        
    // Some terminology lest I lose my mind:
    //  - $file: the array that things like drupal_write_record and cck 
    //      filefield saving will expect. Contains data like fid, title,
    //      alt, nid, filemime and all that jazz; although note that the
    //      fid (which holds a remote ID) has already been unset by now.
    //  - $file_data: the actual raw file.
    //  - $field: the filefield array from CCK, in the event we're a CCK
    //      file. 
    //      TODO: when we come to handle upload files too, this will BREAK! ;)
    $file  = $this->entity;
    $file_data = $this->entity->file;

    // The file is saved differently depending on whether it comes from upload
    // or filefield module.
    // TODO: refactor this for common code.
    switch ($this->data['module']) {
      case 'upload':
        // Create the filepath to save to.
        // Upload files just go in the base files folder.
        $destination_directory = file_directory_path();

        // If a file already existed of the same name it's merely a coincidence, so
        // we use FILE_EXISTS_RENAME to avoid clobbering it.
        $file->filepath = $destination_directory . '/' . $file->filename;
        $filepath = file_save_data($file_data, $file->filepath, FILE_EXISTS_RENAME);

        // Write record in files table.
        drupal_write_record('files', $file);
        $this->id_local = $file->fid;

        // Set the new flag for upload_save() to find... WHICH WON'T WORK.
        // @see upload_transport_process_dependencies_node() for workaround.
        $file->new = TRUE;

        // Uh... who looks at this subsequently?
        $this->data['data']['file'] = $file;

        // The $file->filepath will be wrong in the attaching node's files
        // array, but upload_save() doesn't look at it.
        break;

      case 'filefield':
        $field = $this->data['data']['filefield'];

        // Create the filepath to save to.
        // TODO: brittle! account for upload module too!
        // The way to do this would be for the original entity data array to have a key
        // for originating module, ie upload or filefield.
        // Then here probably invoke a hook to get the right way to get the file's
        // location; though given we only have two cases that may be overkill.
        $destination_directory = filefield_widget_file_path($field);
        // Urgh awful awful functions that both check AND make things simultaneously.
        if (!field_file_check_directory($destination_directory, FILE_CREATE_DIRECTORY)) {
          watchdog('transport', 'Could not create directory %dir to save a file in field %field', array(
            '%dir' => $destination_directory,
            '%field' => $field->name,
          ));
          return;
        }

        $file->filepath = $destination_directory . '/' . $file->filename;

        // ALSO need data that was stored on the CCK filefield when it came over with the node!
        // Do we? what for???
        $filefield_file  = $this->data['data']['file'];

        // Save the incoming file.
        // Gah this is so bloody convoluted. Roll on D7.
        // ALSO! This fails if the local site has not yet saved a file to the
        // current filefield and it has a folder to itself -- because the folder
        // doesn't exist yet!!!!!!
        // If a file already existed of the same name it's merely a coincidence, so
        // we use FILE_EXISTS_RENAME to avoid clobbering it.
        $filepath = file_save_data($file_data, $file->filepath, FILE_EXISTS_RENAME);
        $file->filepath = $filepath;

        // Write record in files table.
        drupal_write_record('files', $file);
        $this->id_local = $file->fid;

        //dsm($file, 'file saved');

        // Put the file back in the data... though DUH we should be using $this->entity,
        // it's what it's there for!
        $this->data['data']['file'] = $file;
        break;
    }
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   File.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_file_d5 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: file - fetch - D5.
   *
   * Fetch a file from a D5 site. 
   * Because there's no file service on D5 Services, so we'd need to backport
   * the module at http://drupal.org/node/290053 and bundle it in to stuff
   * we're providing for D5.
   * In the meantime, the direct approach is fine if all the files are public, 
   * but this needs fixing if they are not: TODO.
   */
  function run_phase_fetch() {
    //dsm($this->object->source);
    
    // Figure out the base URL of the remote site based on
    $remote_url = substr($this->object->source->endpoint, 0, -15);
    
    // Retrieve the filefield data that we sneakily stashed when we came here for dependencies...
    // of course this will completely fail when we're here for upload module files
    // TODO: write a better way. See the TODO at the top of this method.
    $file = $this->object->data['data']['file'];
    
    // Note we have no timestamps coming in from D5.
    $file_url = $remote_url . '/' . $file['filepath'];

    // The URL may have spaces in and PHP is too stupid to figure this out.
    $file_url = str_replace(' ', '%20', $file_url);

    //dsm($file_url);
    $file_data = file_get_contents($file_url);
    if ($file_data === FALSE) {
      // an error happened
    }
    
    // D6 services puts the actual file data here so we will too.
    $file['file'] = $file_data;
    
    // Save the ersatz file array as the remote entity.
    $this->object->entity = (object) $file;
    
    // shortly TODO obsolete
    $this->object->file_data = $file_data;
    //file_put_contents('pic', $file_data);
  }

  /**
   * Phase handler: file - filter - D5
   */
  function run_phase_filter() {
    // Change the  of properties from D5 to D6 style should happen!
    // And obviously, run the alter hook...
    
    //dsm($this->object, 'filter d5');
    $entity_handler = $this->object;
    
    // Filefield on D5 stores title and alt attributes in the base file array;
    // these must be moved to a subarray 'data' for D6.
    $file = $entity_handler->data['data']['file']; // Just a local copy for sanity.
    $entity_handler->data['data']['file']['data'] = array(
      'title' => $file['title'],
      'alt'   => $file['alt'],
    );
    
    // Other things that D6 expects to have...
    $entity_handler->data['data']['file']['list'] = 1; // well this one is for filefield...
    $entity_handler->data['data']['file']['status'] = 1;
    // TODO!!!!
    $entity_handler->data['data']['file']['uid'] = 1;
    // Note also that D5 files don't have a timestamp, but that should be ok.
    
    
    $entity_handler->entity->status = 1;
    // TODO!!!! Set the actual uid from the filefield data!
    $entity_handler->entity->uid = 1;
    
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   File.
 * Source: Clients Drupal 6 Services 2
 */
class TransportRemoteEntityExtender_fetch_file_d6_2 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * The name of the method for run_phase_fetch() to use.
   */
  public $fetch_method_name = 'file.get';

  /**
   * Phase handler: file - filter - D6.
   */
  function run_phase_filter() {
    // Drupal 6 Services encodes file data...
    $this->object->entity->file = base64_decode($this->object->entity->file);
  }

}
