<?php
/**
 * @file transport.remote_entity.node.inc
 * Contains the remote entity class, interfaces, and hooks for nodes.
 *
 * Hook implementations in this file are on behalf of core and contrib modules 
 * providing information about nodes (eg, user module adds the node author as
 * a dependency); they are used by the TransportRemoteEntity_node class and 
 * hence this file is already loaded by that point.
 */

/**
 * Entity class for nodes.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 */
class TransportRemoteEntity_node extends TransportRemoteEntity {
  /**
   * The name of the property that holds the entity's id.
   */
  public $primary_key = 'nid';

  /**
   * Phase handler: node - init.
   */
  function run_phase_init() {
    // It is extremely important to call the parent method here.
    parent::run_phase_init();
  }

  /**
   * Phase handler: primary_key - node.
   *
   * Nodes need to fix up their vid (revision id) as well as their nid.
   * We could call the parent method for the nid but it's easier to see
   * what's going on if we do both here.
   */
  function run_phase_primary_key() {
    $id = $this->primary_key;
    if ($this->local_exists) {
      $this->entity->{$id} = $this->id_local;
      // Load the local node to find out its vid.
      $node_local = node_load($this->id_local);
      $this->entity->vid = $node_local->vid;
    }
    else {
      unset($this->entity->{$id});
      unset($this->entity->vid);
    }
  }

  /**
   * Phase handler: node - save.
   */
  function run_phase_save() {
    //dsm($this->entity, 'about to save node');
    
    // Note that the remote entity's nid and vid properties have already been
    // fixed up so it can be correctly saved whether new or existing locally in
    // the primary_key phase.
    node_save($this->entity);
    $this->id_local = $this->entity->nid;
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Node.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_node_d5 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * The name of the method for run_phase_fetch() to use.
   */
  public $fetch_method_name = 'node.load';

  /**
   * Phase handler: node - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {    
    // It's up to the individual implementation to translate any field names.
    // For instance, if your D5 site has an image in field_foo
    // and you want this to go into foo_bar, you need to move things
    // on the object using hook_transport_TYPE_prefilter_alter() which is
    // invoked in the prefilter phase.

    // Actually very little needs to be done here usually.
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Node.
 * Source: Clients Drupal 6 Services 2
 */
class TransportRemoteEntityExtender_fetch_node_d6_2 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {
    
  /**
   * The name of the method for run_phase_fetch() to use.
   */
  public $fetch_method_name = 'node.get';
  
  /**
   * Phase handler: node - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {
    // Actually very little needs to be done here usually.
  }  
}

// ======================================== Stand-in hook implementations: node

/**
 * Implementation of hook_transport_process_dependencies_node() for user module.
 *
 * Verifies a node for user dependencies from the author of the node.
 *
 * hook_transport_process_dependencies_node() should return an array of
 * dependencies of the given entity, in the form:
 * $dependencies[] = array(
 *  'type' => TYPE,
 *  'id'   => ID,
 *  'ref'  => a variable reference to the actual
 * );
 */
function user_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;

  // The author of the node is a user dependency.
  // If you don't want to bring over users, use the alter hook in the filter phase
  // to change the uid of the node to, say, 1.
  //dsm($remote_entity, 'user deps');
  
  // Users 0 and 1 are never dependencies.
  if ($remote_entity->uid == 0 || $remote_entity->uid == 1) {
    return;
  }

  $deps[] = array(
    'type'  => 'user',
    'id'    => $remote_entity->uid,
    'ref'   => &$remote_entity->uid,
  );
  
  return $deps;
}

/**
 * Implementation of hook_transport_process_dependencies_node() for taxonomy.
 *
 * Adds the node's taxonomy terms as dependencies.
 */
function taxonomy_transport_process_dependencies_node($entity_handler) {
  $remote_node = $entity_handler->entity;

  if (isset($remote_node->taxonomy)) {
    // The XMLRPC call converts objects to arrays, and taxonomy terms should
    // be objects.
    // TODO: move this to some sort of filtering hook...
    $taxonomy = $remote_node->taxonomy;
    unset($remote_node->taxonomy);
    foreach ($taxonomy as $tid => $term) {
      $taxonomy[$tid] = (object) $term;
    }
    $remote_node->taxonomy = $taxonomy;

    // Now add each term as a dependency.
    foreach ($remote_node->taxonomy as $tid => $term) {
      $deps[] = array(
        'type'  => 'term',
        'id'    => $remote_node->taxonomy[$tid]->tid,
        'ref'   => &$remote_node->taxonomy[$tid]->tid,
      );
    }
    return $deps;
  }
}

/**
 * Implementation of hook_transport_process_dependencies_node() for upload.
 *
 * Verifies a node for file dependencies from upload.
 */
function upload_transport_process_dependencies_node($entity_handler) {
  $remote_node = $entity_handler->entity;

  if (isset($remote_node->files)) {
    foreach ($remote_node->files as $fid => $file) {
      $deps[] = array(
        'type' => 'file',
        'id'   => $file['fid'],
        'ref'  => &$remote_node->files[$fid]['fid'],
        // Files are a somewhat special case.
        // upload_save() expects to find $node->files keyed by fid, but this
        // *only* is used when the file is being deleted by the user during node
        // editing. Close shave there!
      );

      // This is ugly. Upload module's upload_save() will only save a record
      // for a file if it finds a $node->files[$fid]['new']. But we
      // can't get data back from the TransportRemoteEntity_file object.
      // So we're going to sneakily check to see if the file has already
      // been retrieved, and set the 'new' flag if not.
      // This works fine, but it'll probably cause pain if you're trying
      // to do dependency alterations around here.
      $file_id_local = transport_get_local_id($entity_handler->source_name, 'file', $file['fid']);
      if (is_null($file_id_local)) {
        $remote_node->files[$fid]['new'] = TRUE;
      }
    }
    return $deps;
  }
}

/**
 * Implementation of hook_transport_process_dependencies_node() for content.
 *
 * Content module does not have any dependencies to add, but we use this
 * hook to add in CCK field data, as it runs before all other CCK field modules.
 *
 * TODO: consider doing this in an earlier phase, eg init?
 */
function content_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
  
  // Stash CCK's field definitions for this node type.
  // Get CCK's data about the local node type and thence a list of its fields.
  $content_type_info = content_types($remote_entity->type);
  $fields = $content_type_info['fields'];
  
  // Restructure this array so it's grouped by module, then by field type.
  // This saves fieldtype modules the bother of having to repeat this sort of thing.
  $fields_by_module = array();
  foreach ($fields as $fieldname => $field) {
    $fields_by_module[$field['module']][$field['type']][$fieldname] = $field;
  }
  //dsm($fields_by_module, 'fields by module');
  // We may get no key for a given CCK module in $fields_by_module
  // causing an error in, eg, nodereference_transport_process_dependencies_node
  // when it tries to iterate. Up to the later hook to cast to array; nothing we can do here.
  
  $entity_handler->module_data['content']['fields'] = $fields_by_module;
}

/**
 * Implementation of hook_transport_process_dependencies_node() for nodereference.
 *
 * Verifies a node for node dependencies from nodereference.
 */
function nodereference_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
  //dsm($remote_entity, 'noderef');
  
  // Retrieve the fieldlist from where content module's hook stored it.
  // We only have one field type anyway, so all our fields are beneath MODULE FIELDTYPE.
  $nodereferencefields = (array) $entity_handler->module_data['content']['fields']['nodereference']['nodereference'];
  // TODO: check field is set here!!!!
  foreach ($nodereferencefields as $field_name => $field) {
    //dsm($remote_entity->$field_name, 'nrf');
    if (isset($remote_entity->$field_name)) {
      foreach ($remote_entity->$field_name as $delta => $data) {
        // An empty noderef field still has a 0 item with a nid key: don't take this!
        if (is_numeric($data['nid'])) {
          $deps[] = array(
            'type'  => 'node',
            'id'    => $data['nid'],
            'ref'   => &$remote_entity->{$field_name}[$delta]['nid'],
          );
        }
      }
    }
  }

  return $deps;
}


/**
 * Implementation of hook_transport_process_dependencies_node() for filefield.
 *
 * Verifies a node for file dependencies in CCK file- and imagefields.
 */
function filefield_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
 
  // Retrieve the fieldlist from where content module's hook stored it.
  // We only have one field type anyway, so all our fields are beneath MODULE FIELDTYPE.
  $filefields = (array) $entity_handler->module_data['content']['fields']['filefield']['filefield'];
  
  // Now we have our data!
  //dsm($remote_entity, 'file deps remote ent');
  //dpr($remote_entity);
  //dsm($filefields);
  
  foreach ($filefields as $field_name => $field) {
    // Check for both existence of the local field on the remote node, and
    // whether there is any data in it; CCK 6 at least has a (mostly) empty 
    // 0 key even if there is no data.
    if (isset($remote_entity->$field_name) && is_array($remote_entity->{$field_name}[0])) {
      foreach ($remote_entity->$field_name as $delta => $data) {
        $deps[] = array(
          'type' => 'file',
          'id'   => $data['fid'],
          'ref'  => &$remote_entity->{$field_name}[$delta]['fid'],
          // Files are a somewhat special case, as to save a file we need the
          // field definition, which by the time we'll be in the remote file
          // entity handler we won't have.
          // So we stash it here.
          // This may well break if the same fid is put into dependencies
          // but with filefield 5 that's REALLY unlikely.
          'data' => array(
            'file'      => $data, 
            'filefield' => $field,
          ),
        );
        
      }      
    }
  }
  
  //dsm($remote_entity, 'done file deps');
  return $deps;  
}

/**
 * Implementation of hook_transport_process_dependencies_node() for filter.
 *
 * Add a node's format property to the list of properties to map.
 * By default, this will not be changed, but can be changed by mappings
 * defined by hook_transport_mappings_list().
 *
 * @see text_transport_process_dependencies_node().
 */
function filter_transport_process_dependencies_node($entity_handler) {
  //dsm($entity_handler);
  $remote_entity = $entity_handler->entity;
  $deps[] = array(
    'type' => 'input_format',
    'id' => $remote_entity->format,
    'ref' => &$remote_entity->format,
    'action' => 'translate',
    'mapping' => 'lookup',
    'lookup group' => 'input_format',
  );
  return $deps;
}

/**
 * Implementation of hook_transport_process_dependencies_node() for (CCK) text.
 *
 * @see filter_transport_process_dependencies_node().
 */
function text_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;

  // Remember that field data stash is grouped by CCK module first, THEN
  // field type to cover the (non-core AFAIK) cases where one module
  // has several types. Hence the slightly WTF double 'text'.
  // The other WTF is that it is essential to cast to an array here.
  $text_fields = (array) $entity_handler->module_data['content']['fields']['text']['text'];
  //dsm($text_fields, 'text fields');

  // From text.module:
  // $options = array(0 => t('Plain text'), 1 => t('Filtered text (user selects input format)'));
  foreach ($text_fields as $field_name => $field) {
    // Only act on text fields set to have an input format, ie not plain text.
    if ($field['text_processing'] == 1) {
      // I wish PHP were smart enough to treat non-existent things as an empty array :/
      if (isset($remote_entity->$field_name) && is_array($remote_entity->$field_name)) {
        foreach ($remote_entity->$field_name as $delta => $data) {
          // We have to add every item to the mapping properties, as each one
          // has its own format.
          $deps[] = array(
            'type' => 'input_format',
            'id' => $remote_entity->{$field_name}[$delta]['format'],
            'ref' => &$remote_entity->{$field_name}[$delta]['format'],
            'action' => 'translate',
            'mapping' => 'lookup',
            'lookup group' => 'input_format',
          );
        }
      }
    }
  }

  return $deps;
}
