<?php
/**
 * @file transport.remote_entity.term.inc
 * Contains the remote entity class, interfaces, and hooks for terms.
 *
 * Requires taxonomy_get_service module on the remote site, part of the
 * taxonomy_services project.
 */

/**
 * Entity class for terms.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 */
class TransportRemoteEntity_term extends TransportRemoteEntity {
  /**
   * The name of the property that holds the entity's id.
   */
  public $primary_key = 'tid';
  
  /**
   * The name of the function to load a local entity.
   * A single parameter if its id is assumed by run_phase_local_load().
   */
  public $entity_load_function = 'taxonomy_get_term';
  // TODO: check this actually gets us everything we need.

  /**
   * Phase handler: term - save.
   */  
  function run_phase_save() {
    //dsm($this->entity, 'about to save term');
      
    // Because taxonomy_save_term() is not a proper API function but a form
    // helper, we have muck about a bit.
    $remote_term = drupal_clone($this->entity);
    $term_array = (array) $remote_term;
    
    // All manner of hoop-jumping to convert the synonyms to the format
    // expected by the function.
    // Also crashes devel module for extra fun!
    // See http://drupal.org/node/1095306.
    $synonyms = $remote_term->synonyms;
    if (count($synonyms) == 1) {
      $s = array_shift($synonyms);
      $term_array['synonyms'] = $s;
    }
    elseif (count($synonyms) > 1) {
      $term_array['synonyms'] = implode("\r\n", $synonyms);
    }

    // Treat parents too...
    $parents = $remote_term->parents;
    if (count($parents) == 1) {
      $single_parent = array_shift($parents);
      $term_array['parent'] = $single_parent;
    }
    else {
      $term_array['parent'] = $parents;
    }

    //dsm($term_array, "about to save term array for " . $this->id_local);
    taxonomy_save_term($term_array);

    // Because we didn't save our actual entity object, we have to retrieve
    // the (potentially) new vid manually.
    $this->entity->tid = $term_array['tid'];

    // Set the local ID on the remote entity object.
    $this->id_local = $this->entity->tid;
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Term.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_term_d5
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * The name of the method for run_phase_fetch() to use.
   */
  public $fetch_method_name = 'taxonomy.loadTerm';

  /**
   * Phase handler: term - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {    
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Term.
 * Source: Clients Drupal 6 Services 2
 */
class TransportRemoteEntityExtender_fetch_term_d6_2 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {
    
  /**
   * The name of the method for run_phase_fetch() to use.
   */
  public $fetch_method_name = 'taxonomy.loadTerm';
  
  /**
   * Phase handler: term - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {    
  }
}

// ======================================== Stand-in hook implementations: term

/**
 * Implementation of hook_transport_process_dependencies_term() for taxonomy module.
 *
 * Here's where it gets a little twisted: this hook allows taxonomy module to state
 * whether an incoming term has dependencies...
 * These are:
 *  - the term's vocabulary
 *  - the term's parents, which are themselves terms
 *  - the term's related terms, which are also terms...
 *
 * hook_transport_process_dependencies_TYPE() should return an array of
 * dependencies of the given entity, in the form:
 * $dependencies[] = array(
 *  'type' => TYPE,
 *  'id'   => ID,
 *  'ref'  => a variable reference to the actual
 * );
 */
function taxonomy_transport_process_dependencies_term($entity_handler) {
  $remote_term = $entity_handler->entity;

  // The term's vocabulary is a dependency.
  // If you don't want to bring over vocabularies, or you want to translate
  // them to existing ones, use the alter hook in the filter phase
  // to change the vid of the term.
  //dsm($remote_entity, 'user deps');
  $deps[] = array(
    'type'  => 'vocabulary',
    'id'    => $remote_term->vid,
    'ref'   => &$remote_term->vid,
  );
  
  // The term's parents are (possibly?) dependencies
  // The term's related terms are probably not dependencies
  // These should probably depend on options.
  // At any rate, they need translating...
  foreach ($remote_term->parents as $tid) {
    $deps[] = array(
      'type'  => 'term',
      'id'    => $remote_term->parents[$tid],
      'ref'   => &$remote_term->parents[$tid],
    );
  }
  
  return $deps;
}
