<?php
/**
 * @file
 * Hooks provided by Transport module.
 */

/**
 * @addtogroup hooks
 * @{
 */

/**
 * Inform the remote entity controller about an entity's dependencies.
 *
 * TYPE is the type of the entity, eg 'node', 'user', etc.
 *
 * @param $entity_handler
 *  The remote entity handler object. Contains the entire retrieved entity
 *  in $entity_handler->entity, as well as other properties pertaining to the
 *  retrieval process.
 *  If you need to store your own data, for example between different hook
 *  implementations, use $entity_handler->module_data[MODULE].
 *  (See content_transport_process_dependencies_node() for an example.)
 *
 * @return
 *  A flat array of dependency items.
 *  Each item is an array itself, which must specify the type and id of the
 *  dependency, and a variable reference to the key holding the id in the
 *  remote entity object. The keys in full are:
 *  - 'type': (required) The entity type, eg 'node', 'file', etc.
 *  - 'id': (required) The entity id.
 *  - 'ref': (required) A variable reference to the property on the entity that stores
 *    the dependency ID. Eg for the node author, this is set to
 *    &$node->uid.
 *    This bit is all quite scary... see the translate phase for more detail.
 *  - 'action': How this dependency should be processed. For normal remote
 *    dependencies, this key can be ignored.
 *    One of:
 *    - 'fetch': (default) The dependency will be added to the list of
 *      entities to retrieve.
 *    - 'translate': The dependency's ID in the current entity is translated
 *      but it is not retrieved.
 *      This is a quick means of avoiding retrieving entities that do not
 *      change, eg files, vocabularies; or entities which should not be
 *      retrieved at all.
 *  - 'mapping': The method to use for translating the remote id. For normal remote
 *    dependencies, this key can be ignored.
 *    One of:
 *      'transport' (default for action 'fetch'): the remote id is mapped
 *        to a local id based on the list of previously retrieved entities.
 *        This allows remote entities that are dependencies to have their
 *        remote ids translated to the local id.
 *      'lookup' (default for action 'translate'): the id is mapped to an
 *        id based on the mapping built from hook_transport_mappings_list().
 *        If this is used, the 'lookup group' key must be set.
 *        This allows you to, for example, translate all input format ids
 *        that pass through Transport, as all entities will have them
 *        as an item to translate.
 *      'fixed': The id is taken directly from the fixed value in 'value',
 *        which obviously must be set.
 *        This is useful for custom hook_transport_TYPE_dependencies_alter()
 *        implementations to force values of dependencies. For example, you
 *        can use this to move terms into a different vocabulary by
 *        implementing hook_transport_term_dependencies_alter() and
 *        setting 'fixed' and the 'value' to the local vocabulary id.
 *  - 'lookup group': Only set this if 'mapping' is set to 'lookup'.
 *  - 'value': Only set this if 'mapping' is set to 'fixed'.
 */
function hook_transport_process_dependencies_TYPE($entity_handler) {
  // Get the remote entity from the handler.
  $remote_entity = $entity_handler->entity;

  // Example for adding node author as a dependency.
  $deps[] = array(
    'type'  => 'user',
    'id'    => $remote_entity->uid,
    'ref'   => &$remote_entity->uid,
  );

  return $deps;
}

/**
 * Alter the list of dependencies returned by modules.
 *
 * This allows you to customize dependencies for your site.
 * Examples in the code below.
 *
 * @param $dependencies
 *  A flat numerical array of all the dependencies as built up by
 *  invoking hook_transport_process_dependencies_TYPE().
 *  In addition to the keys defined there, the items of this array will have:
 *  - 'module': The module which provided this dependency.
 *  Note also that the 'mapping' key will not yet have been set to its default.
 */
function hook_transport_TYPE_dependencies_alter(&$dependency_list) {
  // Example for moving all taxonomy terms into a local vocabulary.
  foreach ($dependency_list as $key => $dependency) {
    if ($dependency['type'] == 'vocabulary') {
      // Act only on terms in vocabulary (remote) 1.
      if ($dependency['id'] == 1) {
        // Change the dependency type from 'fetch' to translate,
        // with a fixed value which we set to vocabulary 2.
        $dependency_list[$key]['action']  = 'translate';
        $dependency_list[$key]['mapping'] = 'fixed';
        $dependency_list[$key]['value'] = 2;
      }
    }
  }
}

/**
 * Provide a list of id mappings.
 *
 * @return
 *  An array of global mappings, in the form:
 *  $mappings[LOOKUP_GROUP] = array(
 *    REMOTE_ID => LOCAL_ID,
 *  );
 * where the LOOKUP_GROUP group corresponds to the 'lookup group' given
 * by hook_transport_process_dependencies_TYPE().
 * (This is not hugely graceful and its only real use case is input formats
 * which are due to eventually be transported as entities too anyway.)
 */
function hook_transport_mappings_list($entity_handler) {
  return array(
    // Map remote input format 3 to local 'Full HTML' format.
    'input_format' => array(
      // remote_format_id => local_format_id,
      3 => 1,
    ),
  );
}

/**
 * Alter the remote entity data ahead of the Transport system's work.
 *
 * This is a very powerful hook which lets you completely modify the remote
 * entity as soon as it has been retrieved from the remote site.
 * You can, for example, move data from one field to another, combine data,
 * rename fields, remove data that you don't want to retrieve.
 * For a previously retrieved entity that is being updated, you can also
 * preserve properties of the local entity by passing their keys to
 * $remote_entity->add_preserve_local(). This for example lets you preserve
 * local taxonomy or node authors while updating the rest of the node.
 *
 * @param $entity_handler
 *  The remote entity handler object. Contains the entire retrieved entity
 *  in $entity_handler->entity, as well as other properties pertaining to the
 *  retrieval process.
 */
function hook_transport_TYPE_prefilter_alter($entity_handler) {
  // Example for setting the node author to uid 1.

  // Get the remote node object.
  $remote_node = $remote_entity->entity;
  $remote_node->uid = 1;
}
