<?php
/**
 * @file transport.controller.inc
 * Contains the transport controller class and subclasses.
 */

/**
 * The transport controller class.
 * 
 * See the transport_demo module for an example of usage.
 * 
 * Create an object of this class to run the transport process for a set of entities.
 * The transport controller needs to be given a source and a list of entities to retrieve. 
 * You can have the controller retrieve entities in small runs, to prevent timeouts. This allows the 
 * controller to be used within a BatchAPI process, or with cron. 
 * It is your responsibility to keep the controller object safe between runs, eg serialized in the database.
 * 
 * Terminology
 * 
 * - Entity: we borrow the Drupal 7 term 'entity' to mean a Drupal object such
 *   as a node, term, user, etc, but extend it to mean any piece of data that 
 *   can be transported; thus a file is treated as an entity (but you may see
 *   it referred to as a pseudoentity, since it has no originating module).
 * - Source: The origin of the data we transport. This is represented by an 
 *   object, usually from Clients module.
 * - Set: A list of entities to retrieve, or the action of retrieving all of these.
 *   One set may include several calls to the controller object's beginRetrieval() method.
 *   So for example, you could call the controller in successive cron runs, and those would
 *   all be part of the same set. You should however preserve the controller object in between
 *   calls: destroying it ends the set.
 *   You may add entities to the set list at any point.
 *   A set is identified by its base time. This is set to be the request time when the controller
 *   is created.
 * 
 * Technology stack
 * 
 * - Application layer: modules can use Transport to create a system for moving content around. Examples
 * include UpRPC which migrates content from old versions of Drupal to new ones, or Content Distribution
 * which synchronizes content between master and slave sites.
 * - Transport: Transport handles fetching a remote entity, including dependencies, and maintaining the relationships
 * between remote and local IDs. Transport layer is what allows the application to merely say: 'fetch nodes 2, 4, 5 from site Foo'.
 * - Clients: Clients module provides a uniform way to connect to remote services. Clients is what allows the 
 * transport layer to say 'get me the result of method node.load on site Foo'.
 * - Services: Services module runs on remote sites and provides the methods to retrieve data. Transport module adds extra services
 * to retrieve data.
 * 
 * Sources
 * 
 * The transporter system is designed to work with Client classes from the Clients module: 
 * one is provided for each version of the Services module. However, it is possible to provide
 * other types of source, since the interaction with a source is implemented in a pluggable interface.
 * Hypothetically you could have a source that just reads data from a file, for example: post a patch! ;)
 * 
 * Dependencies
 * 
 * Each remote entity discovers its dependencies, and the controller object will ensure that no entity is 
 * transported twice in one set.
 * 
 * Execution
 *
 * The controller basically works its way through the stack of entities it's
 * passed initially, but there's a twist!
 * If during the processing of a particular entity, it discovers there are
 * dependencies, then it stops processing, puts the current entity *back on*
 * the stack, then adds the list of dependencies to the stack too.
 * 
 * TODO: Parameters:
 *   - request time (implied)
 *   - entity list
 *   - source type
 *   - source details
 *   - operation_limit
 *   - freshness: the length of time for a local entity to still be fresh. 
 *     Eg, if a node was last retrieved less than a day ago, do not retrieve it again
 *   - operation limit: how many operations the controller should perform before handing back
 *     control to the caller, eg batch operation, cron.
 */
class TransportController {}
class TransportControllerRetriever extends TransportController {
  // TODO: this causes the controller to stop after this many entities have
  // processed. Use for cron runs, batch jobs, etc.
  public $operation_limit;
  
  public $source_name; // The source name.
  public $source; // An object for the source.

  // Lists of entities
  // These are keyed by the unique combination of type and id for easy lookup.
  public $shopping_list = array(); // The list of entities still remaining to retrieve
  public $entity_processed_count; // Count of entities processed so far.
  
  private $entity_tree;
  private $current_entity;
  
  // The time at which we consider the current operation to have started.
  // This is what identifies the current operation uniquely (if you're retrieving
  // from different sources in different processes... you're crazy and you should
  // submit a patch to do this differently.)
  // We guarantee that the same entity will not be retrieved twice since this time.
  private $base_time; 
  
  // Define the phases for the process. A remote entity is at any point in one
  // of these phases, and can go through all or some of them.
  // OBSOLETE
  private $phases = array(
    'init',
    'retrieve',
    'check_local',
    'filter',
    'stash',
    'dependencies',
    'requeue',
    'save',
  );
  
  /**
   * Constructor.
   *
   * @param $source_name
   *  An arbitrary string to uniquely identify a particular source, ie a
   *  particular remote site.
   * @param $source
   *  A source object. This will usually be a client connection class, though
   *  provided the remote entity extenders are defined it could be anything.
   * @param $options
   *  (optional) An associative array with the following keys, all of which are optional:
   *    'operation_limit': the maximum number of operations that should be
   *      carried out by any single call to beginRetrieval(). By default there
   *      is no limit. This needs to exist because what may appear to the
   *      application as a list of 10 node ids could balloon into many more
   *      operations when dependencies get added to the stack.
   *    'base_time': set the base time for the controller instance's operation.
   *      By default this is the server request time when the controller
   *      object is created.
   *    'output': An array containing any of the following options for feedback
   *      and logging as keys with a value of TRUE:
   *      'watchdog': Output to watchdog().
   *      'message': Output to drupal_set_message().
   *      'logfile': Write to a file.
   */
  function __construct($source_name, $source, $options = array()) {
    // Merge in defaults.
    $options += array(
      'operation_limit' => 0,
      'base_time' => $_SERVER['REQUEST_TIME'],
      'output' => array(),
    );

    // The base time is the request time by default, or a given time (for cron runs).
    $this->base_time = $options['base_time'];
    $this->operation_limit = $options['operation_limit'];
    
    $this->source_name  = $source_name;
    $this->source       = $source;
    $this->options      = $options;
    
    // Collect information about remote entity classes.
    $this->remote_entity_classes = module_invoke_all('transport_remote_entity_info');
    //dsm($this->remote_entity_classes);    

    $this->log('Retrieval begun at basetime @time.', array('@time' => $this->base_time));
  }
  
  /**
   * Accessor functions.
   */  
  function get_base_time() {
    return $this->base_time;
  }
  
  function count_entities_remaining() {
    return count($this->shopping_list);
  }

  function count_entities_processed() {
    return $this->entity_processed_count;
  }

  /**
   * Check whether an entity has already been processed by the controller.
   *
   * @param $type
   *  The entity type.
   * @param $id_remote
   *  The entity remote id.
   *
   * @return
   *  TRUE if the entity has been processed by this controller object and is complete.
   */
  function entity_is_processed($type, $id_remote) {
    $result = db_query("SELECT phase FROM {transport_processing_entities} WHERE base_time = %d AND type = '%s' AND id_remote = %d",
      $this->base_time, $type, $id_remote);
    if ($phase = db_result($result)) {
      //dsm("$type, $id_remote - phase $phase");
      if ($phase == 'record') {
        // 'record' is the final phase. Remember to update this if that changes!
        return TRUE;
      }
    }
    return FALSE;
  }
  
  // TODO: Destructor:
  // Empty our stuff in the stash!!
  // -- though what do we mean by 'our'????
  
  /**
   * Add a list of entities to retrieve to the controller's shopping list.
   *
   * This provides an alternative method of operation to the single call
   * to beginRetrieval() when you need to run the set in small chunks. You can
   * call this first with the list of entities, then call beginRetrieval() as 
   * many times as needed with no parameter.
   *
   * @param $entities
   *  The entity IDs to retrieve, grouped into arrays keyed by type.
   *  Example:
   *    $entities = array(
   *      'node' => array(1, 2),
   *      'user' => array(5, 7),
   *    );
   */
  function addEntityList($entities) {
    foreach ($entities as $type => $ids) {
      $class = 'TransportRemoteEntity_' . $type;
      foreach ($ids as $id) {
        // We key the shopping list by the (unique) combination of type and
        // id so we can easily check what's in it.
        $this->shopping_list[$type . '_' .$id] = array(
          'type' => $type,
          'id'   => $id,
          'class' => $class,
        );
      }
    }
    
    //dsm($this->shopping_list, 'shopping list');
  }
  
  /**
   * Entry point to begin retrieval.
   *
   * This can be called as many times as needed with new entities.
   *
   * @param $entities
   *  The entity IDs to retrieve, grouped into arrays keyed by type.
   *  Example:
   *    $entities = array(
   *      'node' => array(1, 2),
   *      'user' => array(5, 7),
   *    );
   *  This is technically optional, because the controller may need to be
   *  several times to work through all dependencies, even once all the
   *  entities to retrieve have been given.
   *  Alternatively, you can pass in a whopping list initially with
   *  addEntityList and then keep calling with no parameter.
   *
   * @return
   *  A list of the entities processed in this particular call.
   *  These are in the same format as the internal lists of entities, that is,
   *  keyed by TYPE_ID and with the keys:
   *    - type
   *    - id
   *    - id_local: the id the entity has received on the local site
   *    - module: the module that added this entity, if a dependency
   *    - class: the class that handled this entity
   *    - ref: magical stuff you shouldn't worry about.
   */
  function beginRetrieval($entities = array()) {
    // Process the list of entities we've been given and add them to end of our list.
    $this->addEntityList($entities);

    // The local process count of how much we've done in this call.
    $process_count = 0;
    
    // This will be populated with the list of what we process this time.
    $current_processed_list = array();
    
    // Work through the list.
    // The shopping_list array may be altered within the loop by entity objects
    // declaring dependencies.
    // We may also bail out of this loop when $process_count reaches a limit.
    while ($entity_data = array_shift($this->shopping_list)) {
      $type = $entity_data['type'];
      $id   = $entity_data['id'];

      // Check whether we've already done this one. If so, skip it.
      if ($this->entity_is_processed($type, $id)) {
        //dsm("$type $id already done, skipping");
        continue;
      }
      
      // Increment the process count no matter what we do in this loop iteration.
      $process_count++;

      // If the entity is stored, that is, if we already began processing it
      // and stalled the process for some reason, retrieve it from the stash.
      if (isset($entity_data['stored'])) {
        //dsm('we already did this one, retrieve it! ' . $type . '_' .$id);
        
        $result = db_query("SELECT object FROM {transport_processing_entities} WHERE base_time = %d AND type = '%s' AND id_remote = %d",
          $this->base_time, $entity_data['type'], $entity_data['id']);
        $remote_entity = unserialize(db_result($result));

        // Restore the values we didn't store.
        $remote_entity->wakeup($this);

        //dsm($remote_entity);
      }
      else {
        // Create a new object of the class for the wanted entity.
        $remote_entity = new $entity_data['class']($this, $entity_data);
      }
      
      //dsm($remote_entity, 'remote entity');
      if (!is_object($remote_entity)) {
        $this->log('No remote entity object constructed for type %type and id %id.', array(
          '%type' => $type,
          '%id'   => $id,
        ));
        // We should return the list processed so far.
        return $current_processed_list;
      }
      
      // Execute the entity handler's main process.
      $remote_entity->run();
      
      // If the entity handler is stalled, deal with its feedback.
      //dsm("status $remote_entity->status");
      switch ($remote_entity->status) {
        case 'error':
          // Entity has raised an error.
          // TODO: handle this better...
          dpm($remote_entity, 'Something has gone badly wrong entity stalled with an error.');
          // TODO: add this to a list of skipped entities.
          // We should return the list processed so far.
          return $current_processed_list;
        case 'skip':
          // Entity has told us to skip it. Do nothing. 
          //dpm($remote_entity, 'we skipped this one.');
          // TODO: add this to a list of skipped entities.
          break;
        case 'dependencies':
          // Set the entity as stored so when we return to it we know to retrieve it.
          $entity_data['stored'] = TRUE;

          //dpm($remote_entity, 'we stopped for dependencies.');
      
          // Add the dependencies to the front of the queue, followed by the
          // current entity so we return to finish its processing.
          // Note that there is no need to check each dependency against the
          // list of those already processed: that will happen when the run
          // loop comes to it in its own course.
          array_unshift($this->shopping_list, $entity_data);
          array_splice($this->shopping_list, 0, 0, $remote_entity->get_fetch_dependencies());
          
          //dsm($this->shopping_list, 'updated shopping list!');
          break;
      }
      
      // TODO: status output depends on our application -- something that runs
      // us on cron will want watchdog, not dsm.
      /*
      drupal_set_message(t('Finished run @count for entity %type @id, status %status, saved as @id-local', array(
        '@count'    => $process_count,
        '%type'     => $entity_data['type'],
        '@id'       => $entity_data['id'],
        '%status'   => $remote_entity->status,
        '@id-local' => $remote_entity->id_local,
        )
      ));
      */
            
      // Record the entity in our list of those processed.
      if ($remote_entity->complete) {
        // Add the local ID to the data in case anything needs it externally.
        $entity_data['id_local'] = $remote_entity->id_local;
        
        // Add the processed entity to both our entire running list and the
        // list for this current run.
        $current_processed_list[$type . '_' .$id] = $entity_data;

        // Increment the total count of processed entities.
        $this->entity_processed_count++;
      }
      
      // Finally, bail if we've hit our operation limit.
      // Note that a value of 0 for the $this->operation_limit setting means 
      // we just keep going until everything is processed.
      if ($this->operation_limit && $process_count >= $this->operation_limit) {
        break;
      }     
    } // while shopping_list
    
    //dsm($current_processed_list, 'processed this cycle');
    return $current_processed_list;
  }

  /**
   * Logging and feedback.
   *
   * Outputs feedback to any of several channels, depending on the options
   * this controller was given on creation.
   *
   * @param $message
   *  The human-readable message. Same as for watchdog().
   * @param $variables
   *  Variables for the message. Same as for watchdog(). Because this may be
   *  output to plain text, % should not be used as a variable prefix.
   * @param $object
   *  An associated object. Not used in all cases.
   * @param $message
   */
  function log($message, $variables = array(), $object = NULL, $severity = WATCHDOG_NOTICE) {
    // Log to Drupal watchdog.
    if ($this->options['output']['watchdog']) {
      watchdog('transport', $message, $variables, $severity);
    }

    // Output a user message.
    if ($this->options['output']['message']) {
      drupal_set_message(t($message, $variables));
    }

    // Output to a logfile.
    // TODO: allow the filename to be set.
    // TODO: remove % in variables!
    if ($this->options['output']['logfile']) {
      $filename = 'transport.log.txt';
      $output = "\n\n" . t($message, $variables);
      if ($object) {
        $output .=  "\n" . print_r($object, TRUE);
      }
      file_put_contents($filename, $output, FILE_APPEND);
    }
  }
}
