<?php
/**
 * @file transport.remote_entity.inc
 * Contains the transport remote entity class, interfaces, and common hooks.
 */

/**
 * The base class to handle a remote entity.
 *
 * The lifecycle of a remote entity runs through several phases.
 * Each phase is run with a standard method, of the form run_phase_PHASENAME().
 * Phase handlers should be small, self-contained methods which do few things,
 * and the emphasis is on letting the base class do as much as possible.
 *
 * Phase methods are defined in the following places:
 *  - the base class
 *  - the class for a particular entity type
 *  - the interface extenders for a particular combination of entity type AND
 *    source type.
 * This does cause a bit of jumping around the code, but with clear commenting
 * it's not too bad ;)
 * This allows us to define a phase method for any combination of entity type
 * and source type. This is what allows different ways to fetch and filter
 * remote data depending on entity type and source type (eg, getting node data
 * from a Drupal 5 site, getting user data from a Drupal 6 site, etc).
 *
 * Each phase may choose to hand over control to the next phase or halt the 
 * entire processing of the entity (and thus return to the controller). It does
 * so by setting the $status property on itself -- see the documentation for
 * possible values and effects.
 *
 * The phases are as follows:
 *
 * - object construction (not really a phase)
 *   - The object is created. Its class reflects the type of entity.
 *   - The object is extended further according to source and type.
 *     TODO: create a system allowing applications to request a subset
 *     of phases, or to add extra phases.
 *   - The local entity is identified, if one exists.
 *   - The stash is checked: if we already fetched this object once in this
 *     set, we are free to skip the fetch phase.
 *     TODO: circular dependencies -- need to think about how to handle these.
 *
 * 'init'
 *    Currently not much, but this is where entity can declare that it should
 *    be skipped. Eg, users with uid 0 and 1 are always skipped.
 *    TODO: at this point we should check to see if the entity has been
 *    previously retrieved within a set timeframe, eg if the node was retrieved
 *    less than a day ago, consider it to still be fresh and skip it.
 *    Freshness should be per-type; eg you might want to say that taxonomy
 *    terms never need to be retrieved again and should always be skipped.
 *    TODO: consider moving things from the constructor to this phase.
 *
 * 'local_load'
 *    Load the local entity if one exists.
 *    The base class assumes a function ENTITY_TYPE_load() exists.
 *    If that's not the case for a particular type, its class should specify a
 *    different function name in the $entity_load_function property.
 *    Alternatively, if you need to do something really wacky, just override
 *    the phase handler method, of course.
 * 
 * 'fetch'
 *    - The data for the entity is fetched from the source.
 *    - The entity is compared with the local version, if it exists.
 *    - TODO: If this is more recent than the required freshness of the
 *      controller operation,
 *      the entity is skipped; control returns to the Controller with a status
 *      to move on to the next entity.
 *
 * 'prefilter'
 *    - The remote entity is passed to hook_transport_TYPE_prefilter_alter().
 *      Modules should use this hook to clean up or change the remote entity
 *      ahead of anything the Transport framework does with it.
 *      Entity classes should probably not override this phase handler.
 *
 * 'filter'
 *   The entity is run through:
 *     1. Custom filters provided by drupal_alter hooks.
 *       This is where you can opt to remove properties that are not relevant to
 *       the local site, or
 *       massage data. 
 *     2. The filter provided by its type, according to source type.
 *       Eg, taxo terms from Drupal 5 automatically get tidied up for the 
 *       current local version of Drupal.
 * 
 * 'dependencies'
 *   The entity is run through a hook that allows all modules to add to a list
 *   of entity dependencies. This is a flat list, eg term 35, node 4, user 3.
 *   At this point, if there are dependencies, execution of the entity handler
 *   stops and control goes back to the controller object. The controller
 *   processes the dependent entities first, and then resumes this entity's
 *   execution at its next phase.
 *   
 * 'translate'   
 *   References in the entity to remote entities are translated to local IDs.
 *   Eg, a node's author is set to a local user (who has been retrieved as a
 *   dependency).
 *
 * 'preexisting' -- TODO
 *   If the entity has not previously been retrieved, check whether an entity
 *   exists locally which matches it anyway. Eg, check for a local user with
 *   the same email address.
 * 
 * 'local_preserve'
 *   Restores requested properties from the existing local entity to the remote
 *   one. In other words, allows specified properties not to be overwritten
 *   when the remote entity is updated.
 *
 * 'save'
 *   The entity is saved.
 *
 * 'record'
 *   The correspondence between remote ID and local ID for this particular
 *   source is stored in the database.
 */
class TransportRemoteEntity extends FacesExtendable {
  // These pretty much all have to be public because objects of this class
  // need to be serialized to the database. Bang goes our privacy :(
  // Hence there's not that much reason to declare them upfront, apart from
  // documentation and trying to prevent key-itis -- listing used object keys
  // may help to spot redundancies and prevent clutter.

  // TODO: phases should get set by the constructor so 
  // more that defined here can be provided.
  // ALSO: need to distinguish phases that take an interface and those that don't!
  // Some phases are commented out and TODO, as I don't need them for this current project.
  public $phases = array(
    'init',
    'local_load',
    'fetch',
    'prefilter',
    'filter',
    'dependencies',
    'translate',
    // 'preexisting', // TODO
    'local_preserve',
    'primary_key',
    'save',
    'record'
  );
  // Phases which need to load interfaces. 
  // Graaargh I don't like the partial duplication with the above $phases, but
  // the alternative is a more complex array. One to ponder. TODO
  // In fact, consider naming faces that are not phase names,
  // as we only need one face in the core module at least.
  // Hence this one face could be called 'core', or 'source' since it
  // diverges on source type.
  public $faced_phases = array(
    'fetch', 
    //'filter',
  );
  
  // Lifecycle properties.
  public $current_phase;   // The phase currently being run or just completed.
  public $phase_checklist; // This array of phases is emptied as each one is run.
  public $status;          // Holds a string describing current status.
    // Possible values include:
    //  - run: handler is running through phases
    //  - error
    //  - complete
    //  - dependencies: handler wants to hand back to the controller for dependencies
    //  - skip: handler wants to hand back to the controller and skip this entity entirely.
  public $advance_phase; // TRUE as long as the handler should continue to work
    // through phases.
    // Phases may set this to FALSE to halt operation.
    // They should ALSO set a status.
    // TODO: phases probably only need to set $this->status to something != 'run'.
    // TODO: refactor this into a method setStatus that's easier (and more rigid) for
    // methods to call, eg setStatus('error'), and the method also takes care of
    // setting advance_phase to FALSE.
  
  // Data about the entity this object handles.
  public $data; // The original data from the controller of what to retrieve.
  public $type;
  public $primary_key; // The name of the property of the entity that holds its id;
    // it's up to subclasses to define this.
  public $id_remote;
  public $id_local; 
  public $entity_load_function; // The name of the function to load the local
    // entity. If not set, TYPE_load() will be assumed by run_phase_local_load().
  public $entity_properties_preserve_local = array(); // 
  
  // The actual entity data.
  public $entity; // This holds the data we fetch from remote
    // and is gradually manipulated over the lifecycle.
  public $result; // This holds exactly what we fetched from remote and is not changed.
  public $entity_local; // The local entity, if it exists.

  public $module_data;
    // Modules may stash anything they like in this array, keyed by module name.
  
  // Data about the local entity.
  public $local_exists = FALSE; // Flag whether we have imported in a previous set. Does not change.
    // If this is TRUE, the $this->id_local also is set.
  public $prior_exists = FALSE; // Flag whether this object exists but has not been retrieved by us.
    // Example: a user with the same email address.

  // Data from the controller: see the controller class for more on these.
  public $controller; // The controller object
  public $base_time;  // The controller's base time of execution. This identifies a set.
  public $source_name;
  public $source;     // The source object. So far, a Clients class.
  
  /**
   * Constructor method.
   *
   * Set properties for the object and loads interfaces.
   */
  function __construct($controller, $entity_data) {
    // Set our controller and copy some data from it.
    $this->controller   = $controller;
    $this->source_name  = $controller->source_name;
    $this->source       = $controller->source;
    $this->base_time    = $controller->get_base_time();
    
    //dsm("basetime: $this->base_time");
    
    // Set our basic parameters from what we were passed in.
    $this->data       = $entity_data;
    $this->type       = $entity_data['type'];
    $this->id_remote  = $entity_data['id']; // BAD NAME. CHANGE.
    
    // Extend the class with interfaces.
    // For each *requested* phase (API TODO!), extend 

    //dsm($this->controller->remote_entity_classes);
    $source_class = get_class($this->source);
  
    // Load the interface for each phase.
    foreach ($this->faced_phases as $phase) {
      //dsm($phase);
      //dsm($this->controller->remote_entity_classes[$this->type]['phases']);
      $interface = 'TransportRemoteEntityInterface_' . $phase;
      $extender_class = $this->controller->remote_entity_classes[$this->type]['phases'][$phase][$source_class];
      //dsm($interface_class);
      $this->extendByClass(array($interface), $extender_class);
    }
    
    // Make a copy of the phase list that we can shift off to keep track of what's next.
    $this->phase_checklist = $this->phases;
    
    // Mark that we're new so drupal_write_record() can be told later
    // when stashing the object between phases.
    $this->new = TRUE;
    
    // Find out if this entity has previously been retrieved.
    $this->get_local_entity();

    // Log that we've started.
    $this->controller->log('Initialized retrieval of @type @id.', array(
      '@type' => $this->type,
      '@id'   => $this->id_remote,
    ));
  }
  
  /**
   * We implement this ourselves to remove the controller object from what is
   * stored in the database, as it's potentially pretty large and doesn't
   * need storing along with this handler.
   * (And there's also a WTF in Faces which requires this to be here.)
   */
  public function __sleep() {
    $array = get_object_vars($this);

    // Remove the controller.
    unset($array['controller']);

    return array_keys($array);
  }

  /**
   * Non-magical wakeup. Restores the controller to this object.
   */
  public function wakeup($controller) {
    $this->controller = $controller;
  }

  /**
   * Accessor functions.
   */
  function get_id() {
    return $this->id_remote;
  }
  
  /**
   * Get the entity's dependency list.
   */
  function get_fetch_dependencies() {
    return $this->dependencies_fetch;
  }

  /**
   * Add keys to the list of those to preserve on the entity.
   *
   * @param $keys
   *  Either an array of key names or a single key name.
   */
  function add_preserve_local($keys) {
    if (!is_array($keys)) {
      $keys = array($keys);
    }

    foreach ($keys as $key) {
      $this->entity_properties_preserve_local[$key] = $key;
    }
  }
  
  /**
   * Get the local id of the remote entity we represent, in the event that
   * it's previously been retrieved.
   */
  function get_local_entity() {
    //dpm("get local $this->source_name, $this->source_id, $this->type");
    $result = db_query("SELECT id FROM {transport_entity_remote} WHERE source = '%s' AND id_remote = %d AND type = '%s'",
      $this->source_name, $this->id_remote, $this->type);
      
    $local_entity_id = (int) db_result($result);
    if ($local_entity_id) {
      $this->id_local     = $local_entity_id;
      $this->local_exists = TRUE;
      //dpm("local exists: $this->type $local_entity_id");
    }
  } 
  
  /**
   * Run the object through as many phases as possible.
   *
   * The controller object calls this. Operation is handed back to it either
   * when the entity is completely retrieved, or it encounters a problem or
   * a dependency to resolve.
   */
  function run() {
    // Loop through phases.
    // Each phase has an entry point function run_phase_FOO.
    // This should set the property $this->advance_phase to:
    //  TRUE if the phase is complete
    //  FALSE otherwise,
    // in which case we need to look into the status of the object for more details.
    $this->advance_phase = TRUE;
    $this->status = 'run';
    //dsm($this->phase_checklist);
    while ($this->advance_phase && count($this->phase_checklist)) {
      
      $this->current_phase = array_shift($this->phase_checklist);
      $this->phase = $this->current_phase; // ugly :(

      // Run the current phase.
      $phase_function = 'run_phase_' . $this->current_phase;
      //dsm("phase $this->current_phase -- function: $phase_function");
      $this->$phase_function();
      
      //$this->advance_phase = FALSE;
      
      // For now, just stash at the end of every phase.
      // TODO: not all phases require stashing; really only those that may halt operation.
      // TODO: bug! If a given entity is a dependency multiple times, then we may
      // get here with the same time, type, id, AND phase and cause a duplicate error 
      // on the database!!!!
      $this->object = $this;
      if (isset($this->new)) {
        drupal_write_record('transport_processing_entities', $this);
        unset($this->new);
      }
      else {
        drupal_write_record('transport_processing_entities', $this, array('base_time', 'type', 'id_remote'));
      }
      unset($this->object);
    } // while loop.
    
    //dsm("left run at phase $this->current_phase status $this->status");
    
    if ($this->status == 'run') {
      // If we've stopped executing phases, but everything is still set to run
      // then it's because we've completed everything.
      if ($this->advance_phase) {
        $this->complete = TRUE;
        $this->status = 'complete';
      }
      else {
        // We should never get here: phases should set a status at the same time
        // as setting advance_phase to FALSE.
        $this->stalled = TRUE;
        $this->status = 'stalled';
      }
    }
  }
  
  /**
   * Phase handler: init - common.
   *
   * Just an empty stub to be overridden for now.
   * TODO: see if anything can be usefully moved here from the contructor method.
   */
  function run_phase_init() {
  }

  /**
   * Phase handler: local_load - common.
   *
   * Load the local entity, if one exists.
   */
  function run_phase_local_load() {
    if ($this->local_exists) {
      // We assume that an entity can be loaded with TYPE_load().
      // If not, classes can override by setting $this->entity_load_function.
      if (!isset($this->entity_load_function)) {
        $this->entity_load_function = $this->type . '_load';
      }

      $function = $this->entity_load_function;
      $this->entity_local = $function($this->id_local);
    }
  }

  /**
   * Phase handler: fetch - (extended in faces).
   *
   * Fetch the object from the remote source and load it into $this->entity.
   *
   * This is implemented in object faces as it differs for each source type.
   * This docblock serves to show where it goes in the phase order for clarity.
   */
  
  /**
   * Phase handler: prefilter - common.
   *
   * Invokes hook_transport_TYPE_prefilter_alter() for modules to alter the
   * entity ahead of the Tranport system's work.
   */
  function run_phase_prefilter() {
    // Invoke hook_transport_TYPE_prefilter_alter().
    $type = $this->type;
    drupal_alter('transport_' . $type . '_prefilter', $this);

    // By now the list of local keys to preserve should be complete
    // (as the modules may set it in the alter hook above).
    // Unset the keys on the entity so the dependencies phase doesn't
    // process them.
    if ($this->local_exists) {
      foreach ($this->entity_properties_preserve_local as $key) {
        unset($this->entity->{$key});
      }
    }
  }

  /**
   * Phase handler: dependencies - common.
   *
   * This should be common to all entities.   
   *
   * Now the object has been filtered we don't need to worry about versions of
   * Drupal: assume it's in a format that could be saved directly.
   *
   * However we may need other objects to exist locally (and have IDs) before
   * we save the current one. If that is the case, this handler is paused and
   * control returns to the controller, which will retrieve and save those,
   * and then return to this handler's execution afterwards.
   *
   * This phase allows modules to add dependencies.
   *
   * TODO: rename this phase (to 'mappings'?) now the 'mappings' phase is
   * obsolete -- this is now general mappings, some of which happen to
   * be dependencies.
   *
   * TODO: add a system to allow the application layer to skip some or
   * all dependencies. Using the filter phase to just delete keys won't
   * cut it, as those may need to be saved anyway.
   * Use case is migration where the order of operations can be fixed in 
   * a logical order.
   * Idea: set a $dep_skip_keys array in the initial $options,
   * and in this phase, stash those keys before invoking the hook,
   * then put them back straight after.
   */  
  function run_phase_dependencies() {
    // Invoke hook_transport_process_dependencies_TYPE() in all modules to
    // let them process the entity and add to its dependency list
    //$dependencies = module_invoke_all('transport_process_dependencies_' . $this->type, $this->entity);
    
    $dependencies_by_module = array();
    $dependencies_hook = 'transport_process_dependencies_' . $this->type;
    foreach (module_implements($dependencies_hook) as $module) {
      $dependencies_by_module[$module] = module_invoke($module, $dependencies_hook, $this);
      // Allow hooks to be lazy and return nothing.
      if (is_null($dependencies_by_module[$module])) {
        $dependencies_by_module[$module] = array();
      }
    }
    //dsm($dependencies_by_module);
    
    // We let implementing modules just return a flat, numeric array where
    // each element has the keys:
    //  - type: the entity type, eg node, user, term
    //  - id: the entity id.
    //  - ref: a variable reference to the property on the entity that stores 
    //      the dependency ID. Eg for the node author, this is set to
    //      &$node->uid.
    //      This bit is all quite scary... see the translate phase for more detail.
    //  - action: whether this is a dependency to retrieve or just translate.
    //    One of:
    //    'fetch' (default): The dependency will be added to the list of
    //      entities to retrieve.
    //    'translate': The dependency's ID in the current entity is translated
    //      but it is not retrieved.
    //      This is a quick means of avoiding retrieving entities that do not
    //      change, eg files, vocabularies; or entities which should not be
    //      retrieved at all.
    //  - mapping: The method to use for translating the remote id.
    //    One of:
    //      'transport' (default for action 'fetch'): the remote id is mapped
    //        to a local id based on the list of previously retrieved entities.
    //      'lookup' (default for action 'translate'): the id is mapped to an
    //        id based on the mapping built from hook_transport_mappings_list().
    //        If this is used, the 'lookup group' key must be set.
    //        This allows you to, for example, translate all input format ids
    //        that pass through Transport, as all entities will have them
    //        as an item to translate.
    //      'fixed': The id is taken directly from the fixed value in 'value',
    //        which obviously must be set.
    //    Tip: you can ignore this key if it scares you! :)
    //  - 'lookup group': Only set this if 'mapping' is set to 'lookup'.
    //  - 'value': Only set this if 'mapping' is set to 'fixed'.
    // This is to provide a nice, uncomplicated (!) API that doesn't involve
    // making your data jump through hoops.

    // Now we rearrange it into a flat list and add defaults.
    // We don't key by TYPE_ID, because we could actually have the same
    // dependency twice but referred to in different places in the entity.
    // Consider for instance the case where two CCK text fields have the same
    // remote input format which must be translated in two places.
    // It's up to get_fetch_dependencies() to collapse down to a shopping list
    // for the controller and use the TYPE_ID key there.
    $dependencies_flat = array();
    foreach ($dependencies_by_module as $module => $module_data) {
      foreach ($module_data as $dep_data) {
        // Add in some extra values and defaults to the dependency data item.
        $dep_data += array(
          'module' => $module,
          'action' => 'fetch',
        );
        // Add the class for dependencies that need fetching
        if ($dep_data['action'] == 'fetch') {
          $dep_data['class'] = 'TransportRemoteEntity_' . $dep_data['type'];
        }

        $dependencies_flat[] = $dep_data;
      }
    }
    //dsm($dependencies_flat, 'flat');

    // Invoke hook_transport_TYPE_dependencies_alter().
    $type = $this->type;
    drupal_alter('transport_' . $type . '_dependencies', $dependencies_flat);
    //dsm($dependencies_flat, 'flat past-alter');

    // Make a separate the list of dependencies into that need to be fetched.
    // This is keyed in the standard form used by the controller class,
    // TYPE_ID.
    // TODO: consider changing the separator to - since that will never
    // occur in machine names!
    $dependencies_fetch = array();
    foreach ($dependencies_flat as $dependency) {
      if ($dependency['action'] == 'fetch') {
        // Create a key in the standard form.
        $key = $dependency['type'] . '_' . $dependency['id'];

        $dependencies_fetch[$key] = $dependency;
      }
    }

    //dsm($dependencies_fetch, 'dependencies to fetch');

    // Store both the list of all dependencies, and that of those that need to be fetched.
    $this->dependencies       = $dependencies_flat;
    $this->dependencies_fetch = $dependencies_fetch;

    // If there are dependencies, stop the process on this entity;
    // return to the controller object.
    if (count($dependencies_fetch)) {
      $this->advance_phase = FALSE;
      $this->status = 'dependencies';
    }
  }

  /**
   * Phase handler: translate - common.
   *
   * Translate any remote ids, including dependencies recorded by the
   * dependencies phase and retrieved during the pause.
   */  
  function run_phase_translate() {
    //dsm($this->dependencies, 'deps translate');
    
    
    // This works by magic.
    // The dependency hooks stored in the deps array a variable reference to the
    // actual variable on the entity that has the remote ID. 
    // We can then just grab this from the deps array and change it, and the entity
    // magically changes too!
    // Eg, we have $deps_key = &$user->uid, and here we change $deps_key.
    // Credit to http://stackoverflow.com/questions/4957232/is-there-way-for-programmatically-storing-an-address-of-properties-for-an-objec
    // Big question though: will it still work if the object is stashed in the database and
    // retrieved on a later run?
    // It seems to!!!!
    foreach ($this->dependencies as $key => $dep) {
      // Add in default value.
      $dep += array(
        'mapping' => ($dep['action'] == 'fetch') ? 'transport' : 'lookup',
      );

      // There are three different ways to get our mapped value.
      switch ($dep['mapping']) {
        case 'transport':
          // Let Transport module translate the id, based on previously retrieved
          // remote entities.
          $id_local  = transport_get_local_id($this->source_name, $dep['type'], $dep['id']);
          //dsm($id_local);
          $this->dependencies[$key]['ref'] = $id_local;
          break;
        case 'lookup':
          // Look up the mapping in an array of data provided by hook_transport_mappings_list().
          //dsm($dep, 'lookup translation');
          $mappings = module_invoke_all('transport_mappings_list', $this);
          //dsm($mappings, 'mappings');
          $group = $dep['lookup group'];
          // Only make the mapping if the mapping list has something.
          if (isset($mappings[$group])) {
            $id_remote = $dep['id'];
            if (isset($mappings[$group][$id_remote])) {
              $this->dependencies[$key]['ref'] = $mappings[$group][$id_remote];
            }
          }
          break;
        case 'fixed':
          // Change the id to a fixed value.
          $this->dependencies[$key]['ref'] = $dep['value'];
          //dsm($dep, 'fixed translation');
          break;
      }
    }
    //$id_local  = transport_get_local_id($entity_handler->source_name, 'user', $remote_entity->uid);
    //$remote_entity->uid = $uid_local;
  }

  /**
   * Phase handler: local_preserve - common.
   *
   * Copy properties from the local entity to the remote one, according to
   * the list set with add_preserve_local().
   * This allows, for example, local taxonomy terms on the node to be preserved
   * when it is updated from remote.
   */
  function run_phase_local_preserve() {
    if ($this->local_exists) {
      foreach ($this->entity_properties_preserve_local as $key) {
        $this->entity->$key = $this->entity_local->$key;
      }
    }
  }

  /**
   * Phase handler: primary_key - common.
   *
   * If the entity already exists locally, put in the correct primary key ID.
   * If it is new locally, unset the primary key ID.
   * This ensures its saving function (and ultimately drupal_write_record()
   * in most cases) correctly updates or inserts accordingly.
   *
   * Entities that have extra work to do here should override this.
   * Eg, node needs to also unset or set $node->vid.
   */
  function run_phase_primary_key() {
    $id = $this->primary_key;
    if ($this->local_exists) {
      $this->entity->{$id} = $this->id_local;
    }
    else {
      unset($this->entity->{$id});
    }
  }

  /**
   * Phase handler: save - (placeholder).
   *
   * This is always overridden; shown here for clarity.
   */
   function run_phase_save() {}

  /**
   * Phase handler: record - common.
   *
   * This should be common to all entities.
   *
   * TODO: only record if something was saved!!!!!
   */
  function run_phase_record() {  
    // Only record if no local id.
    $record = array(
      'source'    => $this->source_name,
      'id_remote' => $this->id_remote,
      'type'      => $this->type,
      'base_time' => $this->base_time,
      'id'        => $this->id_local,
    );
    
    //dsm("recording $this->type $this->id_remote locally as  $this->id_local");
    
    // Test on local_exists, as the id_local will always be set by this point.
    // local_exists represents the state on initialization and does not
    // subsequently change.
    if ($this->local_exists) {
      drupal_write_record('transport_entity_remote', $record, array('source', 'type', 'id_remote'));
    }
    else {
      drupal_write_record('transport_entity_remote', $record);
    }
  }
}

// ======================================== Phase interfaces

/**
 * Some phases the remote entity can go through are handled by an interface.
 
 A thought -- do we really need interfaces for EVERY phase?
 how about just one interface for stuff that is source-specific??
 Let's try sticking it all in the 'fetch' interface until we hit trouble... 
 then if it works, just rename ;)
*/

/**
 * Interface for fetch phase.
 */
interface TransportRemoteEntityInterface_fetch {
  function run_phase_fetch();
  function run_phase_filter();
}

/**
 * Common class for fetch phase. Provides helper functions.
 */
abstract class TransportRemoteEntityExtender_fetch 
  extends FacesExtender 
  implements TransportRemoteEntityInterface_fetch {
    
  /**
   * Phase handler: common - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  /* 
  function run_phase_filter() {
    // This is just some stub code to note that within an interface extender
    // class, the entity is accessed like this:
    // $remote_entity = $this->object->entity;
  }
  */
}

/**
 * Common class for fetch phase on Client sources. Provides helper functions.
 */
abstract class TransportRemoteEntityExtender_fetch_client 
  extends TransportRemoteEntityExtender_fetch
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: common - fetch.
   *
   * Basic handler for the fetch phase with client sources.
   * If the method for your entity type has just an id parameter then you can
   * simply define the variable $fetch_method_name in your class to supply the
   * method name.
   * For more complex requirements, override this in your extender class.
   */
  function run_phase_fetch() {
    // Something in the Services and XMLRPC system is REALLY fussy
    // about strings vs ints, so we have to cast it.
    $remote_id = (int) $this->object->get_id();
    //dsm($this->object);
    // Long chain of objects: because we're in a face, we want the object;
    // and its source is a Clients object on which we call the method to call a method ;)
    try {
      $result = $this->object->source->callMethod($this->fetch_method_name, $remote_id);
    }
    catch (Exception $e) {
      // The clients_connection_drupal_services class throws an exception for
      // XMLRPC errors.
      // Return control to the controller.
      $this->object->advance_phase = FALSE;
      $this->object->status = 'error';
      $this->object->errors['message'] = $e->getMessage();
      // Pass the exception on so the controller can throw it again (TODO).
      $this->object->exception = $e;
      return;
    }

    // Check we got back something sensible. This will handle flagging errors.
    $this->fetch_verify_client_result($result);

    $this->object->result = $result;
    $this->object->entity = (object) $result;

    //dsm($result, 'fetch result');
  }

  /**
   * Check the result of a client XMLRPC method call for sanity.
   *
   * This sets the status to 'error' if there is a problem, and returns
   * flow to the controller.
   */
  function fetch_verify_client_result($result) {
    // Sometimes we just get nothing back at all.
    if (is_null($result)) {
      $this->object->advance_phase = FALSE;
      $this->object->status = 'error';
      $this->object->errors['fetch'] = t('Result of fetch is NULL');
    }
    // TODO: further sanity checks?
  }
}

// ======================================== Stand-in hook implementations: common
/**
 * Implementation of hook_transport_mappings_list() for filter module.
 *
 * @return
 *  An array of global mappings, in the form:
 *  $mappings[GROUP] = array(
 *    REMOTE_ID => LOCAL_ID,
 *  );
 */
/* 
// ARGH: this is really lousy, as it makes absolutely no allowance for 
// different input formats on different remote sites.
// Need to ponder how to handle this for different applications that may
// use transport with several remote sites.
// In the meantime, this is an example of how to get input formats to
// be mapped in your own use of Transport -- implement this hook and 
// return the 'input_format' group as below.
// @see http://drupal.org/node/1064572 for more background.
function filter_transport_mappings_list() {
  return array(
    'input_format' => array(
      // remote_format_id => local_format_id,
      3 => 1,
    ),
  );  
}
*/
